/* OR-2015-01 のための計算実験 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "odicon.h"
#include "con_util.h"

#define DIM 100
#define EPSILON (1e-6)
#define K 20

static double penalty = 1000;
static double lamda[DIM] = { 0.45, 0.55, 0.65, 0.75, 0.85, 0.4, 0.3, 0.4, 0.3, 0.2 };
static double cost[DIM]  = { 500, 70, 93, 25, 500, 30, 20, 10, 10, 5 };
static double alpha = 1; /* for conjugate scaling */


static double F(int k, double lamdaj) {
    int m;
    double all = 0;  /* sum of 0..K (K:constant) */
    double ret = 0;  /* sum of 0..k (k:argument) */

    for (m=0; m<=K; m++) {
        double p = exp(-lamdaj) * pow(lamdaj,m) / factorial(m);
        /*printf("m=%d, lamdaj=%g, p=%g\n", m, lamdaj, p);*/
        all += p;
        if (m <= k) ret += p;
    }
    /* This is not real Poison cumulative distribution function, */
    /* restricted to first K terms in fact. */
    ret /= all;
    /*printf("m=%d, lamdaj=%g, ret=%e\n", m, lamdaj, ret);*/
    assert(0 <= ret);
    assert(fcmp(ret, 1) <= 0);
    return ret;
}

/* cache F() for same lamda[] */
static double cache_F(int x, int j) {
    static double save[1000][DIM];

    assert( 0 <= x );
    assert( x < sizeof(save)/sizeof(*save) );
    assert( 0 <= j );
    assert( j < DIM );
    if (save[x][j] == 0) {
        save[x][j] = F( x, lamda[j] );
    }
    return save[x][j];
}


static double func(int dim, int x[]) {
    int j, k;
    double ret = 0;

    assert( dim <= DIM );
    for (j=0; j<dim; j++) {
        if (x[j] < 0) return PLUS_INFINITY;
    }

    for (k=0; k<=K/**/; k++) {
        double term = 1;
        for (j=0; j<dim; j++) {
            term *= cache_F( x[j]+k, j );  /* same that F( x[j]+k, lamda[j] ) */
        }
        assert(0 <= term);
        assert(term <= 1);
        term = 1 - term;
        ret += term;
        /*if (term < EPSILON) break;*/
    }
    /*printf("k=%d, ret=%g\n", k, ret);*/
    ret *= penalty;

    for (j=0; j<dim; j++) {
        ret += cost[j] * x[j];
    }

    /* ivector_print(dim, x);  printf("; func(x)=%g\n", ret); */
    return ret;
}


static double conjugate(int dim, int y[]) {
    int i;
    static int x[DIM] = { 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 };
    double ret;
    enum verbose_e vl;

    for (i=0; i<dim; i++) cost[i] = alpha * y[i];
    vl = verbose_level_get();
    verbose_level_set(0);
    ret = lgconv_minimize(dim, func, x);
    verbose_level_set(vl);
    return -ret;
}

#ifdef TEST
static void test(int dim) {
    double ans;
    int x[DIM] = {100, 40, 10, 30, 40,10,10,50, 50,50};

    printf("F2(1) = %f\n", F(1, 0.55));
    printf("F2(2) = %f\n", F(2, 0.55));
    printf("F2(3) = %f\n", F(3, 0.55));
    printf("F2(4) = %f\n", F(4, 0.55));

    ans = lgconv_minimize_IFF(dim, func, x);
    if (ans != lgconv_minimize_FW(dim, func, x)) printf("error!\n");
    if (ans != lgconv_minimize(dim, func, x)) printf("error!\n");
}
#endif /* TEST */

static int read_param(int dim) {
    int i;

    if (!fvector_scan(1, &penalty)) return FALSE;
    if (!fvector_scan(dim, cost))   return FALSE;
    if (!fvector_scan(dim, lamda))  return FALSE; /* 1..99 */
    for (i=0; i<dim; i++) lamda[i] /= 100;        /* 0.01 .. 0.99 */
    return TRUE;
}

void demo_inventory2(const int dim) {
    int i, ret;
    int x[DIM];
    int lower[DIM];
    int upper[DIM];
    int sum = 10;

#ifdef TEST
    for (i=0; i<5; i++) printf("%d! = %g\n", i, factorial(i));
    for (i=1; i<200; i++) {
        printf("F2(%d) = %f\n", i, F(i, 0.5));
    }
#endif /* TEST */

    if (dim > DIM) {
        fprintf(stderr, "Too many dimension (%d).\n", dim);
        return;
    }

    for (i=0; i<dim; i++) x[i] = 0;
    if (!read_param(dim)) {
        fprintf(stderr, "Read error.\n");
        return;
    }

    for (i=0; i<dim; i++) { lower[i] = 1; upper[i] = +4; }

    printf("\nprimal ...\n");
    verbose_level_off(STATUS|COUNTER);
    RANGE_PRINT(dim);
    CON_PRINT(dim, func);

    printf("conjugate ...\n");
    RANGE_PRINT(dim);
    CON_PRINT(dim, conjugate);

    for (alpha = 0.1; alpha < 1.3; alpha += 0.2) {
        printf("conjugate alpha = %g ...\n", alpha);
        RANGE_PRINT(dim);
        CON_PRINT(dim, conjugate);
    }
}
