/* This is an implementation of shift scheduling problem
 * by S. Moriguchi and N. Tsuchimura.
 * This program is a part of demonstration CGI.
 * It is called from 'demo-shift.cgi' via 'odicon.cgi'.
 *
 *  c.f. G. Koole and E. van der Sluis (2003): Optimal shift scheduling
 *  with a global service level constraint, IIE Transactions, Vol. 35,
 *  pp. 1049-1055.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "odicon.h"
#include "con_util.h"

#define INTERVAL 100
#define DIM 10

/* determined in demo_shift_schedule() */
static int interval = 13;
static int shift_begin[DIM+1] = {1, 3,  6,  9}; /* 1 origin, including */
static int shift_end[DIM+1]   = {6, 8, 11, 14}; /* 1 origin, excluding */

static double c = 0.01, mu = 20.0;
static double Gr = 100;
static double lambda[INTERVAL] = {88, 87, 86, 85, 55, 1, 1, 1, 1, 1, 1, 1, 1};
static double salary[DIM+1];  /* = {2, 1, 1, 1}; */
/*static int Ds[DIM+1];*/   /* needless Ds*/
static int n_total;
static double alpha = 1; /* for conjugate scaling */

/* needless Ds */
/*
static void D_inv(int x[]){
    int i, sum=0;

    for (i=DIM; i>=0; i--) {
	sum += x[i];
    	salary[i] = sum;
    }
}
*/

/* x^n / n! */
static double exp_per_fact(double x, int n) {
    int i;
    double r = 1.0;

    for (i=n; i>=1; i--) r = r * x / i;
    return r;
}

/* pi = b / (a + b), which is safe even if b=inf */
static double pi(double a, double b) {
    if (b < a) {
        double s = b / a;
	return s / ( 1 + s );
    } else {
        double t = a / b;
	return 1 / ( t + 1 );
    }
}


static double gi(int n, int i, int *outside) {
    double rho, rho_n;
    double a, b, ret;
    int k;

    /*printf("gi: n=%d, i=%d\n", n, i);*/
    rho = lambda[i] / mu;
    rho_n = rho / n;
 
    if (rho_n >=1) {
        *outside = TRUE;
        /*printf("rho_n = %g\n", rho_n);/**/
        return PLUS_INFINITY;
    }
    b = exp_per_fact(rho,n) / (1-rho_n);
    
    a = 0;
    for (k=0; k<n; k++) {
        a += exp_per_fact(rho,k);
    }

    ret = lambda[i] * pi(a, b) * exp(-n * mu*(1-rho_n)*c);

    /*
    printf(" gi[%d] = %g\trho=%g\ta=%g\tb=%g\tpi=%g\texp(*)=%g\n",
	   i, ret, rho, a, b, pi(a, b), exp(-n * mu*(1-rho_n)*c));
    */
    return ret;
}

/* variable translation from L#-convex to multimodular */
static int x2y(int dim, int x[], int y[]) {
    int i, last = 0;

    for (i=0; i<dim; i++) {
        y[i] = x[i] - last;
        last = x[i];
    }
    return last;
}

/* variable translation from multimodular to L#-convex */
static void y2x(int dim, int y[], int x[]) {
    int i, sum = 0;

    for (i=0; i<dim; i++) {
        sum += y[i];
        x[i] = sum;
    }
}


/* func(x) = \sum g_i */
static double func(int dim, int x[]) {
    double t, LAMBDA, s;
    int y[DIM];  /* multimodular */
    int between = 0;
    int outside = FALSE;
    int i, b, e;
  
    assert(dim <= DIM);  /* check dimension */

    y[dim] = n_total - x2y(dim, x, y);
    /*ivector_print(dim, x); printf(" %d\n", n_total);*/
    /*ivector_print(dim+1, y); printf("\n");*/

    for (i=0; i<dim+1; i++) if (y[i] < 0) return PLUS_INFINITY;
    for (i=0; i<INTERVAL; i++) assert(lambda[i] < n_total * mu);

    t = 0;
    b = e = 0;  /* index of shift_*[] */
    for (i=0; i<interval; i++) {
        while (shift_begin[b] == i+1) between += y[b++];
        while (shift_end[e]   == i+1) between -= y[e++];
        /*printf("interval = %d, b=%d, e=%d, #people = %d\n", i+1, b, e, between);*/
        t += gi(between, i, &outside);
        if (outside) return PLUS_INFINITY;
    }

    LAMBDA = 0;
    for (i=0; i<interval; i++) {
        LAMBDA += lambda[i];
    }
    /*printf(" LAMBDA = %g\n\n", LAMBDA);*/

    s = 0;
    for (i=0; i<dim+1; i++) {
        s += salary[i] * y[i];
    }

    if (outside) return PLUS_INFINITY;
    return Gr * t / LAMBDA + s;
}

static int conjugate_x[DIM+1] = { 5,5,11,16 };
static double conjugate(int dim, int y[]) {
    int i, tmp[DIM+1];
    double ret;
    enum verbose_e vl;

    y2x(dim+1, y, tmp);
    for (i=0; i<dim+1; i++) salary[i] = alpha * tmp[i];
    /*printf("salary = "); fvector_print(dim+1, salary); printf("\n");*/
    /*n_total = x[dim];  n_total if given */

    vl = verbose_level_get();
    verbose_level_set(0);
    ret = lgconv_minimize(dim, func, conjugate_x);
    verbose_level_set(vl);
    /*printf("x = "); ivector_print(dim+1, conjugate_x); printf("\n");*/
    return ret;
}

static int read_param(int dim) {
    if (!ivector_scan(1, &interval))       return FALSE;
    if (!ivector_scan(dim+1, shift_begin)) return FALSE;
    if (!ivector_scan(dim+1, shift_end))   return FALSE;

    if (!fvector_scan(1, &c)) return FALSE;
    if (!fvector_scan(1, &mu)) return FALSE;
    if (!fvector_scan(1, &Gr)) return FALSE;
    if (!fvector_scan(interval, lambda)) return FALSE;
    if (!fvector_scan(dim+1, salary)) return FALSE;
    if (!ivector_scan(1, &n_total)) return FALSE;
    return TRUE;
}

void demo_shift_schedule2(const int dim) {
    int y[DIM+1] = {1, 1, 1, 7, 8, 9};  /* solution of multimodular form */
    int x[DIM+1];                 /* initial solution of L# variable form */
    int lower[DIM+1] = { 1, 1, 1, 1 };
    int upper[DIM+1] = { 10, 10, 10, 10 };
    int sum = 20; /* for M-convex/M-concave */
    double ans;
    int i;

#ifdef TEST
    printf("gi = %f\n", gi(5,5));
#endif /* TEST */

    if (dim > DIM) {
        fprintf(stderr, "Illegal dimension (%d).\n", dim);
        return;
    }

    for (i=0; i<dim; i++) x[i] = 3*(i+1);
    if (!read_param(dim)) {
        fprintf(stderr, "Read error.\n");
        return;
    }
    x[dim] = n_total;

    ans = lgconv_minimize(dim, func, x);

    /*printf("S(x) = %f\n", -ans);*/
    printf("\nN = %d\n", n_total);
    x2y(dim+1, x, y);
    printf("\n");
    for (i=0; i<dim+1; i++) {
        printf("x[%d] = %d\n", i+1, x[i]);
    }
    printf("\n");
    for (i=0; i<dim+1; i++) {
        printf("y[%d] = %d\n", i+1, y[i]);
    }

    for (i=0; i<dim; i++) {
        if (dim == 3) { lower[i] = 1; upper[i] = +10; }
        else          { lower[i] = 1; upper[i] =  +5; }
    }
    for (i=0; i<dim; i++) conjugate_x[i] = x[i];

    printf("\nprimal ...\n");
    verbose_level_off(STATUS|COUNTER);
    RANGE_PRINT(dim);
    CON_PRINT(dim, func);

    printf("conjugate ...\n");
    RANGE_PRINT(dim);
    CON_PRINT(dim, conjugate);

    for (alpha = 0.1; alpha < 1.3; alpha += 0.2) {
        printf("conjugate alpha = %g ...\n", alpha);
        RANGE_PRINT(dim);
        CON_PRINT(dim, conjugate);
    }
}
